// tt800.cpp : Defines the exported functions for the DLL application.
//

#include "stdafx.h"

#include <algorithm>
#include <cstdint>

#define export_		extern "C" __declspec(dllexport)
#define calltype	__stdcall

/* Initial Seed Function */

static void init_seed_prng(uint32_t* const p, const int len, uint32_t extraseed)
{
	constexpr uint32_t  mult = 509845221ul;
	constexpr uint32_t  add = 3ul;
	uint32_t        seed1 = 9;
	uint32_t        seed2 = 3402;
	int       i;

	for (i = 0; i < len; i++) {
		seed1 = seed1 * mult + add;
		seed2 *= seed2 + 1;
		extraseed = extraseed * mult + seed2;
		p[i] = (seed2 + (seed1 >> 10)) ^ extraseed;
	}
}

/* TT800 */

class{
	static constexpr uint32_t N = 25, M = 7;
	uint32_t  vektor[N], seed;
	int       idx;

	void update()
	{
		static const uint32_t  A[2] = { 0, 0x8ebfd028 };
		int                    i = 0;

		for (; i < N - M; i++)
			vektor[i] = vektor[i + (M)] ^ (vektor[i] >> 1) ^ A[vektor[i] & 1];
		for (; i < N; i++)
			vektor[i] = vektor[i + (M - N)] ^ (vektor[i] >> 1) ^ A[vektor[i] & 1];
	}

public:
	void Seed(uint32_t s) { seed = s; idx = N + 1; }
	uint32_t GetSeed() { return seed; }
	void SeedAll(const uint32_t* s){ std::copy(s, s+N, vektor); idx = N + 1; }

	uint32_t GetRandom()
	{
		uint32_t e;

		if (idx >= N) {
			if (idx > N)
				init_seed_prng(vektor, N, seed);
			update();
			idx = 0;
		}

		e = vektor[idx++];
		e ^= (e << 7) & 0x2b5b2500UL;             /* Tempering */
		e ^= (e << 15) & 0xdb8b0000UL;
		e ^= (e >> 16);
		return e;
	}
} TT800;


/* Mersenne Twister 19937 */

class{
	static constexpr uint32_t N = 624, M = 397;
	uint32_t  vektor[N], seed;
	int       idx;

	void update()
	{
		static const uint32_t  A[2] = { 0, 0x9908B0DF };
		int                    i = 0;

		for (; i < N - M; i++)
			vektor[i] = vektor[i + (M)] ^ (((vektor[i] & 0x80000000) | (vektor[i + 1] & 0x7FFFFFFF)) >> 1) ^ A[vektor[i + 1] & 1];
		for (; i < N - 1; i++)
			vektor[i] = vektor[i + (M - N)] ^ (((vektor[i] & 0x80000000) | (vektor[i + 1] & 0x7FFFFFFF)) >> 1) ^ A[vektor[i + 1] & 1];
		vektor[N - 1] = vektor[M - 1] ^ (((vektor[N - 1] & 0x80000000) | (vektor[0] & 0x7FFFFFFF)) >> 1) ^ A[vektor[0] & 1];
	}
public:
	void Seed(uint32_t s) { seed = s; idx = N + 1; }
	uint32_t GetSeed() { return seed; }
	void SeedAll(const uint32_t* s) { std::copy(s, s + N, vektor); idx = N + 1; }

	uint32_t GetRandom()
	{
		uint32_t         e;

		if (idx >= N) {
			if (idx > N)
				init_seed_prng(vektor, N, seed);
			update();
			idx = 0;
		}

		e = vektor[idx++];
		e ^= (e >> 11);             /* Tempering */
		e ^= (e << 7) & 0x9D2C5680;
		e ^= (e << 15) & 0xEFC60000;
		e ^= (e >> 18);

		return e;
	}
} MT;


/* Templates */

template<class PRNG>
static inline double rnd(PRNG&& prng)
{
	union { uint64_t i; double f; } v;
	v.i = ((uint64_t) prng.GetRandom()) ^ (((uint64_t)prng.GetRandom()) << 20);
	v.i &= 0x000FFFFFFFFFFFFFULL;
	v.i |= 0x3FF0000000000000ULL;
	return v.f - 1;
}
template<class PRNG>
static inline int32_t rand(PRNG&& prng, int32_t min, int32_t max)
{
	return (int32_t) (rnd(prng) * (max - min + 1) + min);
}

/* TT800 Exported Symbols */

export_ void calltype ttSeed(uint32_t s){ TT800.Seed(s); }
export_ void calltype ttSeedAll(const uint32_t* s) { TT800.SeedAll(s); }
export_ uint32_t calltype ttGetSeed(){ return TT800.GetSeed(); }
export_ uint32_t calltype ttRndInt(){ return TT800.GetRandom(); }
export_ float calltype ttRnd(){ return (float) rnd(TT800); }
export_ int32_t calltype ttRand(int32_t min, int32_t max){ return rand(TT800,min,max); }

/* Mersenne Twister Exported Symbols */

export_ void calltype mtSeed(uint32_t s){ MT.Seed(s); }
export_ void calltype mtSeedAll(const uint32_t* s) { MT.SeedAll(s); }
export_ uint32_t calltype mtGetSeed(){ return MT.GetSeed(); }
export_ uint32_t calltype mtRndInt(){ return MT.GetRandom(); }
export_ float calltype mtRnd(){ return (float)rnd(MT); }
export_ int32_t calltype mtRand(int32_t min, int32_t max){ return rand(MT, min, max); }
